# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *

from .indexing import index_with_searchsorted, index_with_arrays
from .utils import load_arrays, MAX_INT_32, TYPE_DICTIONARY
from scipy import sparse
import numpy as np


class MatrixBuilder(object):
    """
The class, and its subclasses, load structured arrays, manipulate them, and generate `SciPy sparse matrices <http://docs.scipy.org/doc/scipy/reference/sparse.html>`_.

Matrix builders use an array of row indices, an array of column indices, and an array of values to create a `coordinate (coo) matrix <http://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.coo_matrix.html>`_, which is then converted to a `compressed sparse row (csr) matrix <http://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.html>`_.

See the following for more information on structured arrays:

* `NumPy structured arrays <http://docs.scipy.org/doc/numpy/reference/generated/numpy.recarray.html#numpy.recarray>`_
* `Intermediate and processed data <https://docs.brightwaylca.org/intro.html#intermediate-and-processed-data>`_

These classes are not instantiated, and have only `classmethods <https://docs.python.org/2/library/functions.html#classmethod>`__. They are not really true classes, but more organizational. In other words, you should use:

.. code-block:: python

    MatrixBuilder.build(args)

and not:

.. code-block:: python

    mb = MatrixBuilder()
    mb.build(args)

    """

    @classmethod
    def build(cls, paths, data_label,
              row_id_label, row_index_label,
              col_id_label=None, col_index_label=None,
              row_dict=None, col_dict=None, one_d=False, drop_missing=True):
        """
Build a sparse matrix from NumPy structured array(s).

See more detailed documentation at :ref:`building-matrices`.

This method does the following:

TODO: Update

#. Load and concatenate the :ref:`structured arrays files <building-matrices>` in filepaths ``paths`` using the function :func:`.utils.load_arrays` into a parameter array.
#. If not ``row_dict``, use :meth:`.build_dictionary` to build ``row_dict`` from the parameter array column ``row_id_label``.
#. Using the ``row_id_label`` and the ``row_dict``, use the method :meth:`.add_matrix_indices` to add matrix indices to the ``row_index_label`` column.
#. If not ``one_d``, do the same to ``col_dict`` and ``col_index_label``, using ``col_id_label``.
#. If not ``one_d``, use :meth:`.build_matrix` to build a sparse matrix using ``data_label`` for the matrix data values, and ``row_index_label`` and ``col_index_label`` for row and column indices.
#. Else if ``one_d``, use :meth:`.build_diagonal_matrix` to build a diagonal matrix using ``data_label`` for diagonal matrix data values and ``row_index_label`` as row/column indices.
#. Return the loaded parameter arrays from step 1, row and column dicts from steps 2 & 4, and matrix from step 5 or 6.

Args:
    * *paths* (list): List of array filepaths to load.
    * *data_label* (str): Label of column in parameter arrays with matrix data values.
    * *row_id_label* (str): Label of column in parameter arrays with row ID values, i.e. the integer values returned from ``mapping``.
    * *row_index_label* (str): Label of column in parameter arrays where matrix row indices will be stored.
    * *col_id_label* (str, optional): Label of column in parameter arrays with column ID values, i.e. the integer values returned from ``mapping``. Not needed for diagonal matrices.
    * *col_index_label* (str, optional): Label of column in parameter arrays where matrix column indices will be stored. Not needed for diagonal matrices.
    * *row_dict* (dict, optional): Mapping dictionary linking ``row_id_label`` values to ``row_index_label`` values. Will be built if not given.
    * *col_dict* (dict, optional): Mapping dictionary linking ``col_id_label`` values to ``col_index_label`` values. Will be built if not given.
    * *one_d* (bool): Build diagonal matrix.
    * *drop_missing* (bool): Remove rows from the parameter array which aren't mapped by ``row_dict`` or ``col_dict``. Default is ``True``. Advanced use only.

Returns:
    A :ref:`numpy parameter array <building-matrices>`, the row mapping dictionary, the column mapping dictionary, and a COO sparse matrix.

        """
        assert isinstance(paths, (tuple, list, set)), "``paths`` must be a list"
        array = load_arrays(paths)

        if not row_dict:
            row_dict = index_with_searchsorted(
                array[row_id_label],
                array[row_index_label]
            )
        else:
            index_with_arrays(
                array[row_id_label],
                array[row_index_label],
                row_dict
            )

        if one_d:
            # Eliminate references to row data which isn't used;
            # Unused data remains MAX_INT_32 values
            if drop_missing:
                array = array[np.where(array[row_index_label] != MAX_INT_32)]
            matrix = cls.build_diagonal_matrix(array, row_dict, row_index_label, data_label)
        else:
            if not col_dict:
                col_dict = index_with_searchsorted(
                    array[col_id_label],
                    array[col_index_label]
                )
            else:
                index_with_arrays(
                    array[col_id_label],
                    array[col_index_label],
                    col_dict
                )

            if drop_missing:
                array = array[np.where(array[row_index_label] != MAX_INT_32)]
                array = array[np.where(array[col_index_label] != MAX_INT_32)]
            matrix = cls.build_matrix(
                array, row_dict, col_dict, row_index_label, col_index_label,
                data_label)
        return array, row_dict, col_dict, matrix

    @classmethod
    def build_matrix(cls, array, row_dict, col_dict, row_index_label,
                     col_index_label, data_label=None, new_data=None):
        """Build sparse matrix."""
        vector = array[data_label] if new_data is None else new_data
        assert vector.shape[0] == array.shape[0], "Incompatible data & indices"
        # coo_matrix construction is coo_matrix((values, (rows, cols)),
        # (row_count, col_count))
        return sparse.coo_matrix((
            vector.astype(np.float64),
            (array[row_index_label], array[col_index_label])),
            (len(row_dict), len(col_dict))).tocsr()

    @classmethod
    def build_diagonal_matrix(cls, array, row_dict, index_label,
                              data_label=None, new_data=None):
        """Build diagonal sparse matrix."""
        return cls.build_matrix(array, row_dict, row_dict, index_label, index_label, data_label, new_data)


class TechnosphereBiosphereMatrixBuilder(MatrixBuilder):
    """Subclass of ``MatrixBuilder`` that separates technosphere and biosphere parameters"""
    @classmethod
    def build(cls, paths):
        """Build the technosphere and biosphere sparse matrices."""
        assert isinstance(paths, (tuple, list, set)), "paths must be a list"
        array = load_arrays(paths)
        # take ~10 times faster than fancy indexing
        # http://wesmckinney.com/blog/?p=215
        tech_array = array.take(
            np.hstack((
                np.where(array['type'] == TYPE_DICTIONARY["production"])[0],
                np.where(array['type'] == TYPE_DICTIONARY.get("substitution", -999))[0],
                np.where(array['type'] == TYPE_DICTIONARY["technosphere"])[0],
            ))
        )
        bio_array = array.take(np.where(
            array['type'] == TYPE_DICTIONARY["biosphere"]
        )[0])


        activity_dict = index_with_searchsorted(
            tech_array['output'],
            tech_array['col']
        )
        product_dict = index_with_searchsorted(
            tech_array['input'],
            tech_array['row']
        )
        bio_dict = index_with_searchsorted(
            bio_array['input'],
            bio_array['row']
        )
        index_with_arrays(
            bio_array['output'],
            bio_array['col'],
            activity_dict
        )

        technosphere = cls.build_technosphere_matrix(tech_array, activity_dict, product_dict)
        biosphere = cls.build_matrix(bio_array, bio_dict, activity_dict, "row", "col", "amount")
        return (bio_array, tech_array, bio_dict, activity_dict, product_dict,
                biosphere, technosphere)

    @classmethod
    def get_technosphere_inputs_mask(cls, array):
        """Get boolean mask of technosphere inputs from ``array`` (i.e. the ones to include when building the technosphere matrix)."""
        return np.where(
            array["type"] == TYPE_DICTIONARY["technosphere"]
        )

    @classmethod
    def fix_supply_use(cls, array, vector):
        """Make technosphere inputs negative."""
        # Inputs are consumed, so are negative
        mask = cls.get_technosphere_inputs_mask(array)
        vector[mask] = -1 * vector[mask]
        return vector

    @classmethod
    def build_technosphere_matrix(cls, array, activity_dict, product_dict, new_data=None):
        vector = array["amount"] if new_data is None else new_data
        vector = cls.fix_supply_use(array, vector.copy())
        return cls.build_matrix(array, product_dict, activity_dict, "row",
                                "col", "amount", vector)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from .errors import EfficiencyWarning, NoSolutionFound
from .lca import LCA
from scipy.sparse.linalg import lsqr, lsmr
import warnings


class LeastSquaresLCA(LCA):
    """Solve overdetermined technosphere matrix with more products than activities using least-squares approximation.

    See also:

    * `Multioutput processes in LCA <http://chris.mutel.org/multioutput.html>`_
    * `LSMR in SciPy <http://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsmr.html#scipy.sparse.linalg.lsmr>`_
    * `Another least-squares algorithm in SciPy <http://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsqr.html#scipy.sparse.linalg.lsqr>`_

    """
    def solve_linear_system(self, solver=lsmr):
        if self.technosphere_matrix.shape[0] == self.technosphere_matrix.shape[1]:
            warnings.warn("Don't use LeastSquaresLCA for square matrices",
                          EfficiencyWarning)
        self.solver_results = solver(
            self.technosphere_matrix,
            self.demand_array
        )
        if self.solver_results[1] not in {1, 2}:
            warnings.warn(
                "No suitable solution found - supply array is probably nonsense",
                NoSolutionFound
            )
        return self.solver_results[0]

    def decompose_technosphere(self):
        raise NotImplementedError("Can't decompose rectangular technosphere")



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *
import os


class IndepentLCAMixin(object):
    """Mixin that allows `method`, etc. to be filepaths instead of DataStore object names.

    Removes dependency on `bw2data`."""
    def get_array_filepaths(self):
        """Pass through already correct values"""
        assert self.database_filepath, "Must specify `database_filepath` in independent LCA"
        for collection in (self.database_filepath, self.method, self.weighting, self.normalization):
            if collection is not None:
                for fp in collection:
                    assert os.path.exists(fp), "Can't find file {}".format(fp)
        return (
            self.database_filepath,
            self.method,
            self.weighting,
            self.normalization,
        )

    def fix_dictionaries(self):
        """Don't adjust dictionaries even if ``bw2data`` is present, as functional unit is an integer."""
        self._activity_dict = self.activity_dict
        self._product_dict = self.product_dict
        self._biosphere_dict = self.biosphere_dict



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from bw2calc.errors import OutsideTechnosphere, NonsquareTechnosphere
from bw2calc.lca import LCA
from bw2data import *
from bw2data.utils import TYPE_DICTIONARY
from bw2data.tests import BW2DataTest
import numpy as np


class LCACalculationTestCase(BW2DataTest):
    def add_basic_biosphere(self):
        biosphere = Database("biosphere")
        biosphere.register()
        biosphere.write({
            ("biosphere", "1"): {
                'categories': ['things'],
                'exchanges': [],
                'name': 'an emission',
                'type': 'emission',
                'unit': 'kg'
                }})

    def test_basic(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 0.5,
                    'input': ('t', "2"),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", "2"): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.activity_dict[("t", "1")]] = 1
        answer[lca.activity_dict[("t", "2")]] = 0.5
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_redo_lci_fails_if_activity_outside_technosphere(self):
        self.add_basic_biosphere()
        test_data = {("t", "1"): {
            'exchanges': [
                {'amount': 1, 'input': ('biosphere', "1"), 'type': 'biosphere'}
        ]}}
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        more_test_data = {("z", "1"): {
            'exchanges': [
                {'amount': 1, 'input': ('t', "1"), 'type': 'technosphere'}
        ]}}
        more_test_db = Database("z")
        more_test_db.register()
        more_test_db.write(more_test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci()
        with self.assertRaises(OutsideTechnosphere):
            lca.redo_lci({("z", "1"): 1})

    def test_passing_falsey_key(self):
        self.add_basic_biosphere()
        with self.assertRaises(ValueError):
            LCA({None: 1})
        with self.assertRaises(ValueError):
            LCA({(): 1})

    def test_pass_object_as_demand(self):
        self.add_basic_biosphere()

        class Foo:
            pass

        obj = Foo()
        with self.assertRaises(ValueError):
            LCA(obj)

    def test_production_values(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 2,
                    'input': ('t', "1"),
                    'type': 'production',
                    'uncertainty type': 0},
                    {'amount': 0.5,
                    'input': ('t', "2"),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", "2"): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.activity_dict[("t", "1")]] = 0.5
        answer[lca.activity_dict[("t", "2")]] = 0.25
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_substitution(self):
        # bw2data version 1.0 compatibility
        if 'substitution' not in TYPE_DICTIONARY:
            return
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "2"),
                    'type': 'substitution',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", "2"): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.activity_dict[("t", "1")]] = 1
        answer[lca.activity_dict[("t", "2")]] = -1
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_substitution_no_type(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': -1,  # substitution
                    'input': ('t', "2"),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", "2"): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.activity_dict[("t", "1")]] = 1
        answer[lca.activity_dict[("t", "2")]] = -1
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_circular_chains(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 0.5,
                    'input': ('t', "2"),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", "2"): {
                'exchanges': [{
                    'amount': 0.1,
                    'input': ('t', "1"),
                    'type': 'technosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.activity_dict[("t", "1")]] = 20 / 19.
        answer[lca.activity_dict[("t", "2")]] = 10 / 19.
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_only_products(self):
        test_data = {
            ("t", "p1"): {'type': 'product'},
            ("t", "p2"): {'type': 'product'},
            ("t", "a1"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "p1"),
                    'type': 'production',
                }, {
                    'amount': 1,
                    'input': ('t', "p2"),
                    'type': 'production',
                }]
            },
            ("t", "a2"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "p2"),
                    'type': 'production',
                }]
            }
        }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "p1"): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.activity_dict[("t", "a1")]] = 1
        answer[lca.activity_dict[("t", "a2")]] = -1
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_activity_product_dict(self):
        test_data = {
            ("t", "activity 1"): {'type': 'process'},
            ("t", "activity 2"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "activity 2"),
                    'type': 'production',
                }]
            },
            ("t", "activity 3"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "product 4"),
                    'type': 'production',
                }]
            },
            ("t", "product 4"): {
                "type": "product"
            },
        }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "activity 1"): 1})
        lca.lci()
        self.assertEqual(
           [("t", "activity 1"), ("t", "activity 2"), ("t", "activity 3")],
           sorted(lca.activity_dict)
        )
        self.assertEqual(
           [("t", "activity 1"), ("t", "activity 2"), ("t", "product 4")],
           sorted(lca.product_dict)
        )

        ra, rp, rb = lca.reverse_dict()
        self.assertEqual(
           [("t", "activity 1"), ("t", "activity 2"), ("t", "activity 3")],
           sorted(ra.values())
        )
        self.assertEqual(
           [("t", "activity 1"), ("t", "activity 2"), ("t", "product 4")],
           sorted(rp.values())
        )

    def test_process_product_split(self):
        test_data = {
            ("t", "p1"): {'type': 'product'},
            ("t", "a1"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "p1"),
                    'type': 'production',
                }, {
                    'amount': 1,
                    'input': ('t', "a1"),
                    'type': 'production',
                }]
            },
            ("t", "a2"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "p1"),
                    'type': 'production',
                }]
            }
        }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "a1"): 1})
        lca.lci()
        answer = np.zeros((2,))
        answer[lca.activity_dict[("t", "a1")]] = 1
        answer[lca.activity_dict[("t", "a2")]] = -1
        self.assertTrue(np.allclose(answer, lca.supply_array))

    def test_activity_as_fu_raises_error(self):
        test_data = {
            ("t", "p1"): {'type': 'product'},
            ("t", "a1"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "p1"),
                    'type': 'production',
                }, {
                    'amount': 1,
                    'input': ('t', "a1"),
                    'type': 'production',
                }]
            },
            ("t", "a2"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "p1"),
                    'type': 'production',
                }]
            }
        }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        with self.assertRaises(ValueError):
            lca = LCA({("t", "a2"): 1})
            lca.lci()

    def test_nonsquare_technosphere_error(self):
        test_data = {
            ("t", "p1"): {'type': 'product'},
            ("t", "a1"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "p1"),
                    'type': 'production',
                }, {
                    'amount': 1,
                    'input': ('t', "a1"),
                    'type': 'production',
                }]
            },
        }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "a1"): 1})
        with self.assertRaises(NonsquareTechnosphere):
            lca.lci()

    def test_multiple_lci_calculations(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}
                ],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.write(test_data)
        lca = LCA({test_db.random(): 1})
        lca.lci()
        lca.lci()

    def test_dependent_databases(self):
        databases['one'] = {'depends': ['two', 'three']}
        databases['two'] = {'depends': ['four', 'five']}
        databases['three'] = {'depends': ['four']}
        databases['four'] = {'depends': ['six']}
        databases['five'] = {'depends': ['two']}
        databases['six'] = {'depends': []}
        lca = LCA({('one', None): 1})
        self.assertEqual(
            sorted(lca.database_filepath),
            sorted({Database(name).filepath_processed() for name in
                ('one', 'two', 'three', 'four', 'five', 'six')
            })
        )

    def test_filepaths_full(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}
                ],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.write(test_data)
        method = Method(("M",))
        method.register()
        method.write([[("biosphere", "1"), 1.]])
        normalization = Normalization(("N",))
        normalization.register()
        normalization.write([[("biosphere", "1"), 1.]])
        weighting = Weighting("W")
        weighting.register()
        weighting.write([1])
        lca = LCA(
            {("t", "1"): 1},
            method.name,
            weighting.name,
            normalization.name,
        )
        self.assertEqual(
            sorted(Database(x).filepath_processed() for x in ('biosphere', 't')),
            sorted(lca.database_filepath)
        )
        self.assertEqual(
            lca.method_filepath,
            [method.filepath_processed()]
        )
        self.assertEqual(
            lca.weighting_filepath,
            [weighting.filepath_processed()]
        )
        self.assertEqual(
            lca.normalization_filepath,
            [normalization.filepath_processed()]
        )

    def test_filepaths_empty(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}
                ],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        self.assertEqual(
            sorted(Database(x).filepath_processed() for x in ('biosphere', 't')),
            sorted(lca.database_filepath)
        )
        self.assertTrue(lca.method_filepath is None)
        self.assertTrue(lca.normalization_filepath is None)
        self.assertTrue(lca.weighting_filepath is None)

    def test_demand_type(self):
        with self.assertRaises(ValueError):
            LCA(("foo", "1"))
        with self.assertRaises(ValueError):
            LCA("foo")
        with self.assertRaises(ValueError):
            LCA([{"foo": "1"}])

    def test_decomposed_uses_solver(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 0.5,
                    'input': ('t', "2"),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", "2"): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci(factorize=True)
        # Indirect test because no easy way to test a function is called
        lca.technosphere_matrix = None
        self.assertEqual(float(lca.solve_linear_system().sum()), 1.5)

    def test_fix_dictionaries(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 0.5,
                    'input': ('t', "2"),
                    'type': 'technosphere',
                    'uncertainty type': 0},
                    {'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                    'uncertainty type': 0}],
                'type': 'process',
                'unit': 'kg'
                },
            ("t", "2"): {
                'exchanges': [],
                'type': 'process',
                'unit': 'kg'
                },
            }
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)
        lca = LCA({("t", "1"): 1})
        lca.lci()

        supply = lca.supply_array.sum()

        self.assertTrue(lca._fixed)
        self.assertFalse(lca.fix_dictionaries())
        # Second time doesn't do anything
        self.assertFalse(lca.fix_dictionaries())
        self.assertTrue(lca._fixed)
        lca.redo_lci({("t", "1"): 2})
        self.assertEqual(lca.supply_array.sum(), supply * 2)

    def test_basic_lcia(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "2"),
                    'type': 'technosphere',
                }],
            },
            ("t", "2"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                }],
            },
        }
        method_data = [(('biosphere', "1"), 42)]
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)

        method = Method(("a method",))
        method.register()
        method.write(method_data)

        lca = LCA({("t", "1"): 1}, ("a method",))
        lca.lci()
        lca.lcia()

        self.assertTrue(np.allclose(42, lca.score))

    def test_lcia_regionalized_ignored(self):
        test_data = {
            ("t", "1"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('t', "2"),
                    'type': 'technosphere',
                }],
            },
            ("t", "2"): {
                'exchanges': [{
                    'amount': 1,
                    'input': ('biosphere', "1"),
                    'type': 'biosphere',
                }],
            },
        }
        method_data = [
            (('biosphere', "1"), 21),
            (('biosphere', "1"), 21, config.global_location),
            (('biosphere', "1"), 100, "somewhere else"),
        ]
        self.add_basic_biosphere()
        test_db = Database("t")
        test_db.register()
        test_db.write(test_data)

        method = Method(("a method",))
        method.register()
        method.write(method_data)

        lca = LCA({("t", "1"): 1}, ("a method",))
        lca.lci()
        lca.lcia()

        print(lca.score)
        self.assertTrue(np.allclose(42, lca.score))




# -*- coding: utf-8 -*-

__all__ = [
    'DenseLCA',
    'direct_solving_worker',
    'DirectSolvingMixin',
    'DirectSolvingMonteCarloLCA',
    'GraphTraversal',
    'IndepentLCAMixin',
    'LCA',
    'LeastSquaresLCA',
    'load_calculation_package',
    'MatrixBuilder',
    'MonteCarloLCA',
    'MultiLCA',
    'MultiMonteCarlo',
    'ParallelMonteCarlo',
    'ParameterVectorLCA',
    'save_calculation_package',
    'TechnosphereBiosphereMatrixBuilder',
]

__version__ = (1, 6, 1)

from .lca import LCA
from .dense_lca import DenseLCA
from .independent_lca import IndepentLCAMixin
from .least_squares import LeastSquaresLCA
from .multi_lca import MultiLCA
from .graph_traversal import GraphTraversal
from .matrices import MatrixBuilder, TechnosphereBiosphereMatrixBuilder
from .utils import save_calculation_package, load_calculation_package

try:
    from .monte_carlo import (
        direct_solving_worker,
        DirectSolvingMixin,
        DirectSolvingMonteCarloLCA,
        MonteCarloLCA,
        MultiMonteCarlo,
        ParallelMonteCarlo,
    )
    from .mc_vector import ParameterVectorLCA
except ImportError:
    None



import numpy as np
from bw2calc.indexing import index_with_searchsorted, index_with_arrays
import pytest

MAX_INT_32 = 4294967295


def test_index_with_searchsorted():
    inpt = np.array([1, 2, 3, 6, 9, 12, 9, 6, 5])
    expected = np.array([0, 1, 2, 4, 5, 6, 5, 4, 3])
    mapping = {1: 0, 2: 1, 3: 2, 5: 3, 6: 4, 9: 5, 12: 6}
    output = np.zeros(inpt.size)
    result = index_with_searchsorted(inpt, output)
    assert result == mapping
    assert np.allclose(expected, output)

def test_index_with_searchsorted_preserves_dtype():
    inpt = np.array([1, 2, 3, 6, 9, 12, 9, 6, 5])
    output = np.zeros(inpt.size, dtype=np.uint32)
    index_with_searchsorted(inpt, output)
    assert output.dtype == np.uint32

def test_index_with_arrays():
    inpt = np.array([1, 2, 3, 6, 9, 12, 9, 6, 5])
    mapping = {1: 0, 3: 2, 5: 3, 6: 4, 9: 5}
    expected = np.array([0, MAX_INT_32, 2, 4, 5, MAX_INT_32, 5, 4, 3])
    output = np.zeros(inpt.size)
    index_with_arrays(inpt, output, mapping)
    assert np.allclose(output, expected)

def test_index_with_arrays_preserves_dtype():
    inpt = np.array([1, 2, 3, 6, 9, 12, 9, 6, 5])
    mapping = {1: 0, 3: 2, 5: 3, 6: 4, 9: 5}
    output = np.zeros(inpt.size, dtype=np.uint32)
    index_with_arrays(inpt, output, mapping)
    assert output.dtype == np.uint32

def test_index_with_arrays_negative_error():
    inpt = np.array([1, 2, 3, 6, 9, 12, 9, 6, 5])
    mapping = {-1: 0, 3: 2, 5: 3, 6: 4, 9: 5}
    output = np.zeros(inpt.size)
    with pytest.raises(ValueError):
        index_with_arrays(inpt, output, mapping)




# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *

from . import LCA
from heapq import heappush, heappop
import numpy as np
import warnings
try:
    from bw2data import databases
except ImportError:
    databases = {}


class GraphTraversal(object):
    """
Traverse a supply chain, following paths of greatest impact.

This implementation uses a queue of datasets to assess. As the supply chain is traversed, datasets inputs are added to a list sorted by LCA score. Each activity in the sorted list is assessed, and added to the supply chain graph, as long as its impact is above a certain threshold, and the maximum number of calculations has not been exceeded.

Because the next dataset assessed is chosen by its impact, not its position in the graph, this is neither a breadth-first nor a depth-first search, but rather "importance-first".

This class is written in a functional style - no variables are stored in *self*, only methods.

Should be used by calling the ``calculate`` method.

.. warning:: Graph traversal with multioutput processes only works when other inputs are substituted (see `Multioutput processes in LCA <http://chris.mutel.org/multioutput.html>`__ for a description of multiputput process math in LCA).

    """
    def calculate(self, demand, method, cutoff=0.005, max_calc=1e5, skip_coproducts=False):
        """
Traverse the supply chain graph.

Args:
    * *demand* (dict): The functional unit. Same format as in LCA class.
    * *method* (tuple): LCIA method. Same format as in LCA class.
    * *cutoff* (float, default=0.005): Cutoff criteria to stop LCA calculations. Relative score of total, i.e. 0.005 will cutoff if a dataset has a score less than 0.5 percent of the total.
    * *max_calc* (int, default=10000): Maximum number of LCA calculations to perform.

Returns:
    Dictionary of nodes, edges, LCA object, and number of LCA calculations.

        """
        lca, supply, score = self.build_lca(demand, method)
        if score == 0:
            raise ValueError("Zero total LCA score makes traversal impossible")

        # Create matrix of LCIA CFs times biosphere flows, as these don't
        # change. This is also the unit score of each activity.
        characterized_biosphere = np.array((
            lca.characterization_matrix *
            lca.biosphere_matrix).sum(axis=0)).ravel()

        heap, nodes, edges = self.initialize_heap(
            demand, lca, supply, characterized_biosphere)
        nodes, edges, counter = self.traverse(
            heap, nodes, edges, 0, max_calc, cutoff, score, supply,
            characterized_biosphere, lca, skip_coproducts)

        return {
            'nodes': nodes,
            'edges': edges,
            'lca': lca,
            'counter': counter,
        }

    def initialize_heap(self, demand, lca, supply, characterized_biosphere):
        """
Create a `priority queue <http://docs.python.org/2/library/heapq.html>`_ or ``heap`` to store inventory datasets, sorted by LCA score.

Populates the heap with each activity in ``demand``. Initial nodes are the *functional unit*, i.e. the complete demand, and each activity in the *functional unit*. Initial edges are inputs from each activity into the *functional unit*.

The *functional unit* is an abstract dataset (as it doesn't exist in the matrix), and is assigned the index ``-1``.

        """
        heap, edges = [], []
        nodes = {-1: {
            'amount': 1,
            'cum': lca.score,
            'ind': 1e-6 * lca.score
        }}
        for activity_key, activity_amount in demand.items():
            index = lca.activity_dict[activity_key]
            cum_score = self.cumulative_score(
                index, supply, characterized_biosphere, lca
            )
            heappush(heap, (abs(1 / cum_score), index))
            nodes[index] = {
                "amount": float(supply[index]),
                "cum": cum_score,
                "ind": self.unit_score(index, supply, characterized_biosphere)
            }
            edges.append({
                "to": -1,
                "from": index,
                "amount": activity_amount,
                "exc_amount": activity_amount,
                "impact": cum_score * activity_amount / float(supply[index]),
            })
        return heap, nodes, edges

    def build_lca(self, demand, method):
        """Build LCA object from *demand* and *method*."""
        lca = LCA(demand, method)
        lca.lci()
        lca.lcia()
        lca.decompose_technosphere()
        return lca, lca.solve_linear_system(), lca.score

    def cumulative_score(self, index, supply, characterized_biosphere, lca):
        """Compute cumulative LCA score for a given activity"""
        demand = np.zeros((supply.shape[0],))
        demand[index] = supply[index]
        return float((characterized_biosphere * lca.solver(demand)).sum())

    def unit_score(self, index, supply, characterized_biosphere):
        """Compute the LCA impact caused by the direct emissions and resource consumption of a given activity"""
        return float(characterized_biosphere[index] * supply[index])

    def traverse(self, heap, nodes, edges, counter, max_calc, cutoff,
                 total_score, supply, characterized_biosphere, lca,
                 skip_coproducts):
        """
Build a directed graph by traversing the supply chain.

Node ids are actually technosphere row/col indices, which makes lookup easier.

Returns:
    (nodes, edges, number of calculations)

        """
        static_databases = {name for name in databases if databases[name].get('static')}
        reverse, _, _ = lca.reverse_dict()

        while heap:
            if counter >= max_calc:
                warnings.warn("Stopping traversal due to calculation count.")
                break
            parent_index = heappop(heap)[1]
            # Skip links from static databases
            if static_databases and reverse[parent_index][0] in static_databases:
                continue

            # Assume that this activity produces its reference product
            scale_value = lca.technosphere_matrix[parent_index, parent_index]
            if scale_value == 0:
                raise ValueError(u"Can't rescale activities that produce "
                                 u"zero reference product")
            col = lca.technosphere_matrix[:, parent_index].tocoo()
            # Multiply by -1 because technosphere values are negative
            # (consumption of inputs) and rescale
            children = [(int(col.row[i]), float(-1 * col.data[i] / scale_value))
                for i in range(col.row.shape[0])]
            for activity, amount in children:
                # Skip values on technosphere diagonal
                if activity == parent_index:
                    continue
                # Skip negative coproducts
                if skip_coproducts and amount <= 0:
                    continue
                counter += 1
                cumulative_score = self.cumulative_score(
                    activity, supply, characterized_biosphere, lca)
                if abs(cumulative_score) < abs(total_score * cutoff):
                    continue
                # Edge format is (to, from, mass amount, cumulative impact)
                edges.append({
                    "to": parent_index,
                    "from": activity,
                    # Amount of this link * amount of parent demanding link
                    "amount": amount * nodes[parent_index]["amount"],
                    # Raw exchange value
                    "exc_amount": amount,
                    # Amount of dataset demanded
                    "impact": float(amount * nodes[parent_index]["amount"]
                    # times impact per unit of this input
                    * cumulative_score / supply[activity])
                })
                # Want multiple incoming edges, but don't add existing node
                if activity in nodes:
                    continue
                nodes[activity] = {
                    # Total amount of this flow supplied
                    "amount": supply[activity],
                    # Cumulative score from all flows of this activity
                    "cum": cumulative_score,
                    # Individual score attributable to environmental flows
                    # coming directory from or to this activity
                    "ind": self.unit_score(activity, supply,
                                           characterized_biosphere)
                }
                heappush(heap, (abs(1 / cumulative_score), activity))

        return nodes, edges, counter



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *

from .lca import LCA
from numpy.linalg import solve


class DenseLCA(LCA):
    def solve_linear_system(self):
        """
Master solution function for linear system :math:`Ax=B`.

    To most numerical analysts, matrix inversion is a sin.

    -- Nicolas Higham, Accuracy and Stability of Numerical Algorithms, Society for Industrial and Applied Mathematics, Philadelphia, PA, USA, 2002, p. 260.

We use `UMFpack <http://www.cise.ufl.edu/research/sparse/umfpack/>`_, which is a very fast solver for sparse matrices.

If the technosphere matrix has already been factorized, then the decomposed technosphere (``self.solver``) is reused. Otherwise the calculation is redone completely.

        """
        return solve(self.technosphere_matrix.todense(), self.demand_array)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
from eight import *

from .utils import MAX_INT_32
import itertools
import numpy as np


def index_with_arrays(array_from, array_to, mapping):
    """Map ``array_from`` keys to ``array_to`` values using the dictionary ``mapping``.

    Turns the keys and values of mapping into index arrays.

    This is needed to take the ``flow``, ``input``, and ``output`` columns, which can be arbitrarily large integers, and transform them to matrix indices, which start from zero.

    Here is an example:

    .. code-block:: python

        import numpy as np
        a_f = np.array((1, 2, 3, 4))
        a_t = np.zeros(4)
        mapping = {1: 5, 2: 6, 3: 7, 4: 8}
        index_with_arrays(a_f, a_t, mapping)
        # => a_t is now [5, 6, 7, 8]

    Args:
        * *array_from* (array): 1-dimensional integer numpy array.
        * *array_to* (array): 1-dimensional integer numpy array.
        * *mapping* (dict): Dictionary that links ``mapping`` indices to ``row`` or ``col`` indices, e.g. ``{34: 3}``.

    Operates in place. Doesn't return anything."""
    keys = np.array(list(mapping.keys()))
    values = np.array(list(mapping.values()))

    if keys.min() < 0:
        raise ValueError("Keys must be positive integers")

    index_array = np.zeros(keys.max() + 1) - 1
    index_array[keys] = values

    mask = array_from <= keys.max()
    array_to[:] = -1
    array_to[mask] = index_array[array_from[mask]]
    array_to[array_to == -1] = MAX_INT_32


def index_with_searchsorted(array_from, array_to):
    """Build a dictionary from the sorted, unique elements of an array, and map this dictionary from ``array_from`` to ``array_to``.

    Adapted from http://stackoverflow.com/questions/3403973/fast-replacement-of-values-in-a-numpy-array.

    Here is an example:

    .. code-block:: python

        import numpy as np
        array = np.array((4, 8, 6, 2, 4))
        output = np.zeros(5)
        index_with_searchsorted(array, output)
        # => returns {2: 0, 4: 1, 6: 2, 8: 3}
        # and `output` is [1, 3, 2, 0, 1]

    ``array_from`` and ``array_to`` are arrays of integers.

    Returns a dictionary that maps the sorted, unique elements of ``array_from`` to integers starting with zero."""
    unique = np.unique(array_from)
    idx = np.searchsorted(unique, array_from)
    array_to[:] = idx
    return dict(zip(
        (int(x) for x in unique),
        itertools.count()
    ))



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *
from future.utils import implements_iterator

from .lca import LCA
from .utils import clean_databases
from bw2data import projects
from contextlib import contextmanager
from scipy.sparse.linalg import iterative
try:
    from pypardiso import spsolve
except ImportError:
    from scipy.sparse.linalg import spsolve
from stats_arrays.random import MCRandomNumberGenerator
import multiprocessing
import sys

if sys.version_info < (3, 0):
    # multiprocessing.pool as a context manager not available in Python 2.7
    @contextmanager
    def pool_adapter(pool):
        try:
            yield pool
        finally:
            pool.terminate()
else:
    pool_adapter = lambda x: x


@implements_iterator
class IterativeMonteCarlo(LCA):
    """Base class to use iterative techniques instead of `LU factorization <http://en.wikipedia.org/wiki/LU_decomposition>`_ in Monte Carlo."""
    def __init__(self, demand, method=None, iter_solver=iterative.cgs,
                 seed=None, *args, **kwargs):
        super(IterativeMonteCarlo, self).__init__(demand, method=method, *args,
                                                  **kwargs)
        self.seed = seed
        self.iter_solver = iter_solver
        self.guess = None
        self.lcia = method is not None

    def __iter__(self):
        return self

    def __call__(self):
        return next(self)

    def __next__(self):
        raise NotImplemented

    def solve_linear_system(self):
        if not self.iter_solver or self.guess is None:
            self.guess = spsolve(
                self.technosphere_matrix,
                self.demand_array)
            return self.guess
        else:
            solution, status = self.iter_solver(
                self.technosphere_matrix,
                self.demand_array,
                x0=self.guess,
                maxiter=1000)
            if status != 0:
                return spsolve(
                    self.technosphere_matrix,
                    self.demand_array
                )
            return solution


class DirectSolvingMixin(IterativeMonteCarlo):
    def solve_linear_system(self):
        return LCA.solve_linear_system(self)


@implements_iterator
class MonteCarloLCA(IterativeMonteCarlo):
    """Monte Carlo uncertainty analysis with separate `random number generators <http://en.wikipedia.org/wiki/Random_number_generation>`_ (RNGs) for each set of parameters."""
    def load_data(self):
        self.load_lci_data()
        self.tech_rng = MCRandomNumberGenerator(self.tech_params, seed=self.seed)
        self.bio_rng = MCRandomNumberGenerator(self.bio_params, seed=self.seed)
        if self.lcia:
            self.load_lcia_data()
            self.cf_rng = MCRandomNumberGenerator(self.cf_params, seed=self.seed)
        if self.weighting:
            self.load_weighting_data()
            self.weighting_rng = MCRandomNumberGenerator(self.weighting_params, seed=self.seed)

    def __next__(self):
        if not hasattr(self, "tech_rng"):
            self.load_data()
        self.rebuild_technosphere_matrix(self.tech_rng.next())
        self.rebuild_biosphere_matrix(self.bio_rng.next())
        if self.lcia:
            self.rebuild_characterization_matrix(self.cf_rng.next())
        if self.weighting:
            self.weighting_value = self.weighting_rng.next()

        if not hasattr(self, "demand_array"):
            self.build_demand_array()

        self.lci_calculation()
        if self.lcia:
            self.lcia_calculation()
            if self.weighting:
                self.weighting_calculation()
            return self.score
        else:
            return self.supply_array


class DirectSolvingMonteCarloLCA(MonteCarloLCA, DirectSolvingMixin):
    pass


@implements_iterator
class ComparativeMonteCarlo(IterativeMonteCarlo):
    """First draft approach at comparative LCA"""
    def __init__(self, demands, *args, **kwargs):
        self.demands = demands
        # Get all possibilities for database retrieval
        demand_all = demands[0].copy()
        for other in demands[1:]:
            demand_all.update(other)
        super(ComparativeMonteCarlo, self).__init__(demand_all, *args, **kwargs)

    def load_data(self):
        self.load_lci_data()
        self.load_lcia_data()
        self.tech_rng = MCRandomNumberGenerator(self.tech_params, seed=self.seed)
        self.bio_rng = MCRandomNumberGenerator(self.bio_params, seed=self.seed)
        self.cf_rng = MCRandomNumberGenerator(self.cf_params, seed=self.seed)

    def __next__(self):
        if not hasattr(self, "tech_rng"):
            self.load_data()
        self.rebuild_technosphere_matrix(self.tech_rng.next())
        self.rebuild_biosphere_matrix(self.bio_rng.next())
        self.rebuild_characterization_matrix(self.cf_rng.next())

        results = []
        for demand in self.demands:
            self.build_demand_array(demand)
            self.lci_calculation()
            self.lcia_calculation()
            results.append(self.score)
        return results


def single_worker(args):
    project, demand, method, iterations = args
    projects.set_current(project, writable=False)
    mc = MonteCarloLCA(demand=demand, method=method)
    return [next(mc) for x in range(iterations)]


def direct_solving_worker(args):
    project, demand, method, iterations = args
    projects.set_current(project, writable=False)
    mc = DirectSolvingMonteCarloLCA(demand=demand, method=method)
    return [next(mc) for x in range(iterations)]


class ParallelMonteCarlo(object):
    """Split a Monte Carlo calculation into parallel jobs"""
    def __init__(self, demand, method, iterations=1000, chunk_size=None,
                 cpus=None):
        clean_databases()
        self.demand = demand
        self.method = method
        self.cpus = cpus or multiprocessing.cpu_count()
        if chunk_size:
            self.chunk_size = chunk_size
            self.num_jobs = iterations // chunk_size
            if iterations % self.chunk_size:
                self.num_jobs += 1
        else:
            self.num_jobs = self.cpus
            self.chunk_size = (iterations // self.num_jobs) + 1

    def calculate(self, worker=single_worker):
        with pool_adapter(multiprocessing.Pool(processes=self.cpus)) as pool:
            results = pool.map(
                worker,
                [
                    (projects.current, self.demand, self.method, self.chunk_size)
                    for _ in range(self.num_jobs)
                ]
            )
        return [x for lst in results for x in lst]


def multi_worker(args):
    """Calculate a single Monte Carlo iteration for many demands.

    ``args`` are in order:
        * ``project``: Name of project
        * ``demands``: List of demand dictionaries
        * ``method``: LCIA method

    Returns a list of results: ``[(demand dictionary, result)]``

    """
    project, demands, method = args
    projects.set_current(project, writable=False)
    lca = LCA(demands[0], method)
    lca.load_lci_data()
    lca.load_lcia_data()
    # Create new matrices
    lca.rebuild_technosphere_matrix(next(
        MCRandomNumberGenerator(lca.tech_params)))
    lca.rebuild_biosphere_matrix(next(
        MCRandomNumberGenerator(lca.bio_params)))
    lca.rebuild_characterization_matrix(
        next(MCRandomNumberGenerator(lca.cf_params)))
    lca.lci(factorize=True)
    lca.lcia()
    results = []
    for demand in demands:
        lca.redo_lcia(demand)
        results.append((demand, lca.score))
    return results


class MultiMonteCarlo(object):
    """
This is a class for the efficient calculation of *many* demand vectors from
each Monte Carlo iteration.

Args:
    * ``args`` is a list of demand dictionaries
    * ``method`` is a LCIA method
    * ``iterations`` is the number of Monte Carlo iterations desired
    * ``cpus`` is the (optional) number of CPUs to use

The input list can have complex demands, so ``[{('foo', 'bar'): 1, ('foo', 'baz'): 1}, {('foo', 'another'): 1}]`` is OK.

Call ``.calculate()`` to generate results.

    """
    def __init__(self, demands, method, iterations, cpus=None):
        clean_databases()
        # Convert from activity proxies if necessary
        self.demands = [{(k[0], k[1]): v}
                        for obj in demands
                        for k, v in obj.items()]
        self.method = method
        self.iterations = iterations
        self.cpus = cpus or multiprocessing.cpu_count()

    def merge_results(self, objs):
        """Merge the results from each ``multi_worker`` worker.

        ``[('a', [0,1]), ('a', [2,3])]`` becomes ``[('a', [0,1,2,3)]``.

        """
        r = {}
        for obj in objs:
            for key, value in obj:
                r.setdefault(frozenset(key.items()), []).append(value)
        return [(dict(x), y) for x, y in r.items()]

    def calculate(self, worker=multi_worker):
        """Calculate Monte Carlo results for many demand vectors.

        Returns a list of results with the format::

            [(demand dictionary, [lca scores])]

        There is no guarantee that the results are returned in the same order as the ``demand`` input variable.

        """
        with pool_adapter(multiprocessing.Pool(processes=self.cpus)) as pool:
            results = pool.map(
                worker,
                [
                    (projects.current, self.demands, self.method)
                    for _ in range(self.iterations)
                ]
            )
        return self.merge_results(results)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from eight import *
from future.utils import implements_iterator

from .monte_carlo import IterativeMonteCarlo
from .utils import extract_uncertainty_fields as euf
from stats_arrays.random import MCRandomNumberGenerator
import numpy as np


@implements_iterator
class ParameterVectorLCA(IterativeMonteCarlo):
    """A Monte Carlo class where all uncertain parameters are stored in a single large array.

    Useful for sensitivity analysis and easy manipulation."""
    def load_data(self):
        self.load_lci_data()
        positions = {
            "tech": (0, self.tech_params.shape[0]),
            "bio": (
                self.tech_params.shape[0],
                self.tech_params.shape[0] + self.bio_params.shape[0]
            )
        }

        # `euf` is extract_uncertainty_fields; needed because we are
        # "gluing" together arrays with different column numbers and labels
        params = (euf(self.tech_params), euf(self.bio_params))

        if self.lcia:
            self.load_lcia_data()
            positions["cf"] = (
                positions["bio"][1],
                positions["bio"][1] + self.cf_params.shape[0]
            )
            params = params + (euf(self.cf_params),)

        if self.weighting:
            self.load_weighting_data()
            positions["weighting"] = (
                positions["bio"][1],
                positions["bio"][1] + self.cf_params.shape[0]
            )
            params = params + (euf(self.weighting_params),)

        self.positions = positions
        self.params = np.hstack(params)
        self.rng = MCRandomNumberGenerator(self.params, seed=self.seed)

    def rebuild_all(self, vector=None):
        """Rebuild the LCI/LCIA matrices from a new Monte Carlo sample or provided vector."""
        if not hasattr(self, "positions"):
            self.load_data()

        if vector is not None and not isinstance(vector, np.ndarray):
            raise ValueError("`vector` must be a 1-d numpy array")

        if vector is not None:
            assert vector.shape == self.params.shape, \
                "Incorrect `vector` shape. Is {}, but should be {}".format(
                    vector.shape, self.params.shape
                )
        # Copy to break references and avoid later manipulation by RNG
        self.sample = (self.rng.next() if vector is None else vector).copy()
        self.rebuild_technosphere_matrix(self.tech_sample)
        self.rebuild_biosphere_matrix(self.bio_sample)
        if self.lcia:
            self.rebuild_characterization_matrix(self.cf_sample)
        if self.weighting:
            self.weighting_value = self.weighting_sample

    def __next__(self):
        """Generate a new Monte Carlo iteration."""
        self.rebuild_all()

        if not hasattr(self, "demand_array"):
            self.build_demand_array()

        self.lci_calculation()
        if self.lcia:
            self.lcia_calculation()
            if self.weighting:
                self.weighting_calculation()
            return self.score
        else:
            return self.supply_array

    @property
    def tech_sample(self):
        return self.sample[self.positions["tech"][0]:self.positions["tech"][1]]

    @property
    def bio_sample(self):
        return self.sample[self.positions["bio"][0]:self.positions["bio"][1]]

    @property
    def cf_sample(self):
        return self.sample[self.positions["cf"][0]:self.positions["cf"][1]]

    @property
    def weighting_sample(self):
        return self.sample[self.positions["weighting"][0]:self.positions["weighting"][1]]
